package com.vcarecity.publish.controller;

import com.vcarecity.publish.annotation.SkipMethod;
import com.vcarecity.publish.api.ApiResult;
import com.vcarecity.publish.api.ApiStatus;
import com.vcarecity.publish.constants.ElasticConstant;
import com.vcarecity.publish.elastic.service.ElasticLoadDataService;
import com.vcarecity.publish.pojo.dto.TotalResultDTO;
import com.vcarecity.publish.pojo.dto.UnitAgencyDTO;
import com.vcarecity.publish.pojo.dto.UnitIdDTO;
import com.vcarecity.publish.pojo.query.UnitAgencyQuery;
import com.vcarecity.publish.sql.service.UnitService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.io.IOException;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @author Kerry on 19/12/11
 */

@Slf4j
@RestController
public class MainController implements PrepareController {

    private AtomicInteger dataStatus;

    private final UnitService unitService;
    private final ElasticLoadDataService elasticLoadDataService;

    public MainController(UnitService unitService,
                          ElasticLoadDataService elasticLoadDataService,
                          @Value("${app.scheduling.enable}") Boolean schedulingEnable) {
        this.unitService = unitService;
        this.elasticLoadDataService = elasticLoadDataService;

        if (schedulingEnable) {
            dataStatus = new AtomicInteger(ElasticConstant.DATA_STATUS_LOADED);
        } else {
            dataStatus = new AtomicInteger(ElasticConstant.DATA_STATUS_NOT_LOAD);
        }
    }


    @ResponseBody
    @RequestMapping(value = "/load", method = {RequestMethod.GET, RequestMethod.POST})
    public ApiResult loadData() {
        synchronized (this) {
            if (dataStatus.get() == ElasticConstant.DATA_STATUS_NOT_LOAD) {
                logger.info("prepare to load data into elasticsearch");
                dataStatus.set(ElasticConstant.DATA_STATUS_LOADING);
                new Thread(() -> {
                    try {
                        elasticLoadDataService.loadData();
                        dataStatus.set(ElasticConstant.DATA_STATUS_LOADED);
                    } catch (IOException e) {
                        e.printStackTrace();
                        dataStatus.set(ElasticConstant.DATA_STATUS_NOT_LOAD);
                    }
                }).start();
                return ApiResult.builder().status(40000).message("first load data into elasticsearch").build();
            } else if (dataStatus.get() == ElasticConstant.DATA_STATUS_LOADING) {
                return ApiResult.builder().status(40000).message("loading data into elasticsearch").build();
            }
        }
        return ApiResult.builder().status(40000).message("already load data success!").build();
    }


    @PostMapping("/query-unit-agency")
    public ApiResult queryUnitAgency(@RequestBody @Valid UnitAgencyQuery unitAgencyQuery) throws IOException {

        TotalResultDTO<List<UnitIdDTO>> totalResult = unitService.getUnitAgencyByQuery(unitAgencyQuery, UnitIdDTO.class, ElasticConstant.UNIT_AGENCY_INDEX);

        return ApiResult.builder()
                .status(ApiStatus.SUCCESS_CODE)
                .data(totalResult.getData())
                .total(totalResult.getTotal())
                .build();
    }

    @PostMapping("/test/query-unit-agency")
    public ApiResult testQueryUnitAgency(@RequestBody @Valid UnitAgencyQuery unitAgencyQuery) throws IOException {
        TotalResultDTO<List<UnitAgencyDTO>> totalResult = unitService.getUnitAgencyByQuery(unitAgencyQuery, UnitAgencyDTO.class, ElasticConstant.UNIT_AGENCY_INDEX);
        return ApiResult.builder()
                .status(ApiStatus.SUCCESS_CODE)
                .data(totalResult.getData())
                .total(totalResult.getTotal())
                .build();
    }


    @SkipMethod
    @Override
    public int getDataStatus() {
        return dataStatus.get();
    }
}
