package com.vcarecity.publish.sql.service.impl;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.vcarecity.elastic.util.AgencyPathUtil;
import com.vcarecity.publish.pojo.dto.TotalResultDTO;
import com.vcarecity.publish.pojo.dto.UnitIdDTO;
import com.vcarecity.publish.pojo.query.UnitAgencyQuery;
import com.vcarecity.publish.sql.service.UnitService;
import com.vcarecity.publish.util.ObjectFieldToMap;
import com.vcarecity.publish.util.Utils;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.*;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.sort.SortOrder;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * @author Kerry on 19/12/11
 */

@Slf4j
@Service
public class UnitServiceImpl implements UnitService {

    private final RestHighLevelClient restHighLevelClient;
    private final ObjectMapper objectMapper;

    static final String[] SEARCH_KEY_NAMES = new String[]{
            "agencyName.keyword",
            "agencyDetail.keyword",
            "unitType.keyword",
            "nameOfBuilding.keyword",
    };

    static final String[] MATCH_KEY_NAMES = new String[]{
            "agencyAttributeClassId",
            "regulatoryLevelId",
            "unitTypeId",
            "property",
            "classify"
    };


    public UnitServiceImpl(RestHighLevelClient restHighLevelClient,
                           ObjectMapper objectMapper) {
        this.restHighLevelClient = restHighLevelClient;
        this.objectMapper = objectMapper;
    }

    @Override
    public <T extends UnitIdDTO> TotalResultDTO<List<T>> getUnitAgencyByQuery(UnitAgencyQuery queryParam, Class<T> cls, String... indices) throws IOException {


        List<Integer> agencyIdList = AgencyPathUtil.splitAgencyId(queryParam.getAgencyPath());
        if (agencyIdList.isEmpty()) {
            return new TotalResultDTO<>(Collections.emptyList(), 0L);
        }

        String[] queryFields = ObjectFieldToMap.getField(cls);

        QueryBuilder agencyIdBoolQuery = agencyIdListCondition(agencyIdList);
        QueryBuilder keywordBoolQuery = multiInputParamKeywordBool(queryParam.getKeyword());

        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery().must(agencyIdBoolQuery);

        if (keywordBoolQuery != null) {
            queryBuilder.must(keywordBoolQuery);
        }

        List<QueryBuilder> queryBuilders = arrayParamQuery(queryParam);
        for (QueryBuilder builder : queryBuilders) {
            queryBuilder.must(builder);
        }

        int from = (queryParam.getPage() - 1) * queryParam.getSize();

        SearchRequest searchRequest = new SearchRequest();
        SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();
        searchSourceBuilder.query(queryBuilder);
        searchSourceBuilder.fetchSource(queryFields, new String[]{});
        searchSourceBuilder.from(from);
        searchSourceBuilder.size(queryParam.getSize());

        searchSourceBuilder.sort("property", SortOrder.ASC);
        searchSourceBuilder.sort("unitUpdateStamp", SortOrder.ASC);
        searchSourceBuilder.sort("unitId", SortOrder.ASC);

        searchRequest.source(searchSourceBuilder);
        searchRequest.indices(indices);


        logger.debug("getUnitAgencyByQuery:{}", searchSourceBuilder);

        SearchResponse searchResponse = restHighLevelClient.search(searchRequest, RequestOptions.DEFAULT);

        final SearchHits hits = searchResponse.getHits();

        long total = hits.getTotalHits().value;

        List<T> res = new ArrayList<>(hits.getHits().length);


        for (SearchHit hit : hits.getHits()) {
            T t = objectMapper.readValue(hit.getSourceAsString(), cls);
            res.add(t);
        }
        return new TotalResultDTO<>(res, total);
    }

    /**
     * agency id
     * <pre>
     * {
     *   "size" : 100,
     *   "query" : {
     *     "bool" : {
     *       "must" : [
     *         {
     *           "bool" : {
     *             "must" : [
     *               {
     *                 "term" : {
     *                   "aId0" : {
     *                     "value" : 1,
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               },
     *               {
     *                 "term" : {
     *                   "aId1" : {
     *                     "value" : 3,
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               }
     *             ],
     *             "adjust_pure_negative" : true,
     *             "boost" : 1.0
     *           }
     *         },
     *         {
     *           "term" : {
     *             "aId2" : {
     *               "value" : 67921,
     *               "boost" : 1.0
     *             }
     *           }
     *         }
     *       ],
     *       "adjust_pure_negative" : true,
     *       "boost" : 1.0
     *     }
     *   },
     *   "_source" : {
     *     "includes" : [
     *       "unitId"
     *     ],
     *     "excludes" : [ ]
     *   },
     *   "sort" : [
     *     {
     *       "_doc" : {
     *         "order" : "asc"
     *       }
     *     }
     *   ]
     * }
     * </pre>
     *
     * @param agencyIds
     * @return
     */
    private QueryBuilder agencyIdListCondition(List<Integer> agencyIds) {

        if (agencyIds == null || agencyIds.size() == 0) {
            return null;
        }

        boolean isOnlyOnceId = agencyIds.size() == 1;

        final BoolQueryBuilder boolQueryBuilder = QueryBuilders.boolQuery();

        final Map<String, Integer> map = AgencyPathUtil.agencyPathQueryMap(agencyIds);

        for (Map.Entry<String, Integer> entry : map.entrySet()) {
            TermQueryBuilder termQueryBuilder = QueryBuilders.termQuery(entry.getKey(), entry.getValue());
            if (isOnlyOnceId) {
                return termQueryBuilder;
            }
            boolQueryBuilder.must(termQueryBuilder);
        }
        return boolQueryBuilder;
    }


    /**
     * key word
     * <pre>
     *  {
     *   "size" : 100,
     *   "query" : {
     *     "bool" : {
     *       "must" : [
     *         {
     *           "bool" : {
     *             "must" : [
     *               {
     *                 "bool" : {
     *                   "must" : [
     *                     {
     *                       "term" : {
     *                         "aId0" : {
     *                           "value" : 1,
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     },
     *                     {
     *                       "term" : {
     *                         "aId1" : {
     *                           "value" : 3,
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     }
     *                   ],
     *                   "adjust_pure_negative" : true,
     *                   "boost" : 1.0
     *                 }
     *               },
     *               {
     *                 "term" : {
     *                   "aId2" : {
     *                     "value" : 67921,
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               }
     *             ],
     *             "adjust_pure_negative" : true,
     *             "boost" : 1.0
     *           }
     *         },
     *         {
     *           "bool" : {
     *             "should" : [
     *               {
     *                 "bool" : {
     *                   "should" : [
     *                     {
     *                       "wildcard" : {
     *                         "agencyName.keyword" : {
     *                           "wildcard" : "*鱼峰区*",
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     },
     *                     {
     *                       "wildcard" : {
     *                         "agencyDetail.keyword" : {
     *                           "wildcard" : "*鱼峰区*",
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     }
     *                   ],
     *                   "adjust_pure_negative" : true,
     *                   "boost" : 1.0
     *                 }
     *               },
     *               {
     *                 "wildcard" : {
     *                   "nameOfBuilding.keyword" : {
     *                     "wildcard" : "*鱼峰区*",
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               }
     *             ],
     *             "adjust_pure_negative" : true,
     *             "boost" : 1.0
     *           }
     *         }
     *       ],
     *       "adjust_pure_negative" : true,
     *       "boost" : 1.0
     *     }
     *   },
     *   "_source" : {
     *     "includes" : [
     *       "unitId"
     *     ],
     *     "excludes" : [ ]
     *   },
     *   "sort" : [
     *     {
     *       "_doc" : {
     *         "order" : "asc"
     *       }
     *     }
     *   ]
     * }
     * </pre>
     *
     * @param keyword
     * @return
     */
    private BoolQueryBuilder multiInputParamKeywordBool(String keyword) {
        if (keyword == null || keyword.trim().length() == 0) {
            return null;
        }


        final BoolQueryBuilder boolQueryBuilder = QueryBuilders.boolQuery();

        for (String kn : SEARCH_KEY_NAMES) {
            QueryBuilder queryBuilder = QueryBuilders.wildcardQuery(kn, "*" + keyword + "*");
            // QueryBuilder queryBuilder = QueryBuilders.matchQuery(kn, keyword);
            boolQueryBuilder.should(queryBuilder);
        }
        return boolQueryBuilder;
    }


    private List<QueryBuilder> arrayParamQuery(UnitAgencyQuery queryParam) {
        List<QueryBuilder> queryArray = new ArrayList<>(MATCH_KEY_NAMES.length);

        for (String matchKeyName : MATCH_KEY_NAMES) {

            Object value = null;
            try {
                Field declaredField = UnitAgencyQuery.class.getDeclaredField(matchKeyName);
                declaredField.setAccessible(true);
                value = declaredField.get(queryParam);

            } catch (NoSuchFieldException | IllegalAccessException e) {
                e.printStackTrace();
            }
            if (value == null) {
                continue;
            }
            String listString = value.toString();
            int[] ints = Utils.paramToArray(listString);
            if (ints.length == 1) {
                TermQueryBuilder termQueryBuilder = QueryBuilders.termQuery(matchKeyName, ints[0]);
                queryArray.add(termQueryBuilder);
            } else if (ints.length > 1) {
                TermsQueryBuilder termsQueryBuilder = QueryBuilders.termsQuery(matchKeyName, ints);
                queryArray.add(termsQueryBuilder);
            }

        }

        return queryArray;
    }

}
